/*
Copyright (c) 2013 V.Ayllón.  All rights reserved.
 
 DiyouPCBRun is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 DiyouPCBRun, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with Microayl so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import processing.serial.*;
import java.awt.event.*;

float scaler = 1;
float nmx, nmy, nmz= 0.;
float u, v = 0.;
boolean locked = false;
float ejxOffset, xOffset = 0.0; 
float ejyOffset, yOffset = 0.0; 


PGraphics pg;

color blueRay = color(100, 0, 255);
color grayRay = color(184, 134, 11);
color myColor = grayRay;

boolean ON = true;
boolean OFF = false;
boolean laser_state = OFF;
boolean stopped = false;
long unknowCmd = 0l;
long maxErrors = 5;

int X_MAX_PRT = 212;
int X_MIN_PRT = 0;
int Y_MAX_PRT = 170;
int Y_MIN_PRT = 0;

int SCALE = 4;
int X_AXIS = 0;
int Y_AXIS = 1;
int NUM_AXIS = 2;

int X_MAX_POS = 180*SCALE;
int X_MIN_POS = 0*SCALE;
int Y_MAX_POS = 180*SCALE;
int Y_MIN_POS = 0*SCALE;

int X_MAX_LENGTH = (X_MAX_POS - X_MIN_POS);
int Y_MAX_LENGTH =(Y_MAX_POS - Y_MIN_POS);

int XTRA_WIDTH = 200;  // Extra length for command window and controls
int XTRA_LENGTH = 150;  // Extra length for msg window and top gui

int WINDOW_LENGTH = Y_MAX_LENGTH+XTRA_LENGTH;
int WINDOW_WIDTH  = X_MAX_LENGTH+XTRA_WIDTH;

int X_MARGIN = 5;
int Y_MARGIN = 5;

int[] DRAW_WINDOW_COORDS = {
  X_MARGIN, 120, X_MAX_LENGTH+(X_MARGIN*2), 120+Y_MAX_LENGTH+(Y_MARGIN*2)
  };

int[] GRAPH_WINDOW_SIZE = {
  DRAW_WINDOW_COORDS[2]-DRAW_WINDOW_COORDS[0], DRAW_WINDOW_COORDS[3]-DRAW_WINDOW_COORDS[1], 85
};

float[] offset =  new float[NUM_AXIS];
float[] current_position = new float[NUM_AXIS];
float[] destination = new float[NUM_AXIS];
float feedrate = 0;
float feedmultiply = 1;

boolean paused = false;
String loadPath = null;
BufferedReader fileReader = null;
String buffer = null;
String fileRow[] = null;
int rowNum = 0;

Serial myPort;        // The serial port
int autoState;        // Automata state

String myString;

/**
 ** setup()
 **/
void setup() {
  imageMode(CORNERS);
  size(WINDOW_LENGTH, WINDOW_WIDTH, P3D);
  pg = createGraphics(GRAPH_WINDOW_SIZE[0], GRAPH_WINDOW_SIZE[1], P3D);
  noStroke();
  cp5 = new ControlP5(this);
  background(0);
  GUIinit();
  HPGLInit();
  autoState = 0;
  addMouseWheelListener(new MouseWheelListener() { 
    public void mouseWheelMoved(MouseWheelEvent mwe) { 
      mouseWheel(mwe.getWheelRotation());
    }
  }
  ); 
  background(128);
}


/**
 ** Draw() - This is the main thread
 **/
void draw() {
  smooth();
  pg.smooth();
  imageMode(CORNERS);

  switch(autoState) {
  case 0:    // Waiting state
    delay(10);
    if (!portStatus)
      connectCtrl.setImage(imgs[4]);
    break;
  case 1: //  Send buffer
    myColor = blueRay;
    if ((buffer = sendBuffer()) == null) {
      autoState = 0;  // EOF
    } 
    else {
      if (processCommands(buffer)) {
        autoState = 3;  // Error
      } 
      else {
        autoState = 2;  // Wait CTS
      }
    }
    break;
  case 2:    //  Waiting CTS state
    while (myPort.available () > 0) {
      myString = myPort.readStringUntil('\n');
      if (myString != null) {
        if (codeSeen(myString, "ok")) {
          autoState = 4;
        } else if (codeSeen(myString, "error")) {
            autoState = 6;
          } else {
            printlnCmd(myString);
          }
        }
      }
      break;
    case 3:    // File Error state, probably the file is not HPGL
      printlnMsg("Too many parsing errors. May be the file is not HPGL.");
      printlnCmd("-- Aborted --");
      resetThis(1);
      autoState = 0;
      break;
    case 4:    // Simulation state, only print the file on screen
      pg.beginDraw();
      pg.background(0);  // Clean the window
      pg.translate((GRAPH_WINDOW_SIZE[0]/2) + nmx * scaler-100, (GRAPH_WINDOW_SIZE[1]/2) + nmy*scaler - 100, (GRAPH_WINDOW_SIZE[2]/2) + nmz*scaler - 100);
      pg.scale(scaler);
      pg.rotateX(u);
      pg.rotateY(v);
      pg.translate(-(GRAPH_WINDOW_SIZE[0]/2), -(GRAPH_WINDOW_SIZE[1]/2), -(GRAPH_WINDOW_SIZE[2]/2));
      // Redraw
      for (int i = 0 ; i < fileRow.length; i++) {
        if (rowNum == 0) {  // not started print
          myColor = grayRay;
        } 
        else {              // Printing or paused
          if (i < rowNum)
            myColor = blueRay;
          else
            myColor = grayRay;
        }
        if (processCommands(fileRow[i])) {
          autoState = 3;  // Error
          break;
        }
      }
      pg.endDraw();

      if (autoState == 4) {  //OK, End state
        image(pg, DRAW_WINDOW_COORDS[0], DRAW_WINDOW_COORDS[1], DRAW_WINDOW_COORDS[2], DRAW_WINDOW_COORDS[3]);
        if (printStatus==1)
          autoState = 1;      // Printing
        else
          autoState = 0;      // Wait
      }
      break;
    case 5:    // Waiting file openning
      if (loadPath != null) {
        fileRow = null;
        if (openFile()) {
          autoState = 3;  // Error
        }
        else {
          rowNum = 0;
          autoState = 4;  // Load image
        }
      } 
      else {
        delay(10);
        autoState = 5;  // Load image
      }
      break;
    case 6:    // Firmware Error
      printlnCmd("-- Aborted --");
      resetThis(1);
      autoState = 0;
      break;
    default:
      break;
    }
  }

  /**
   **processCommands() - Process a command from the buffer
   **/
  boolean processCommands(String buf) 
  { 
    int command = 0;

    if (buf == null)
      return false;

    stringToChar(buf);

    command = HPGLParserRun(cmdbuffer);

    switch(command) {
    case 0x5055: //PU:
      laser_state = OFF;
      //    println("Pen up"); 
      break;
    case 0x5044: //PD:
      laser_state = ON;
      //    println("Pen down"); 
      break;
    case 0x5041: //PA:
      if (stopped == false) {
        numpar[0] = numpar[0] + DRAW_WINDOW_COORDS[0] + 40;
        numpar[1] = DRAW_WINDOW_COORDS[3]-numpar[1];
        setCoordinates(); 
        prepareMove();
      }
      break;
    case 0x4141: //AA:
      if (stopped == false) {
        numpar[0] = numpar[0] + DRAW_WINDOW_COORDS[0] + 40;
        numpar[1] = DRAW_WINDOW_COORDS[3]-numpar[1];
        prepareArcMove();
      }
      break;
    case 0x494E: //IN:		// Home & init
      // home
      homeAllAxis();
      break;
    case 0x5653: //VS:
      setFeedRate(numpar[0]);
      break;
    case 0x5057: //PW:
      break;
    case 0x4048: //@H
    case 0x4049: //@I
    case 0x5350: //SP:		// Select pen 
    case 0x4950: //IP:		// Init point
    case 0x5343: //SC:		// Scale all of them processed inside parser
      break;
    default:
      unknowCmd++;
      if (unknowCmd > maxErrors)
        return true;
      break;
    }
    return false;
  }

  /**
   ** codeSeen()
   **/
  boolean codeSeen(String s, String code)
  {
    String[] m1 = match(s, code);
    return (m1 != null);
  }
  /**
   ** stringToChar()
   **/
  void stringToChar(String s) {
    int i = 0;
    for (i = 0; i < s.length(); i++)
      cmdbuffer[i] = s.charAt(i);
    cmdbuffer[i] = '\0';
  }

  /**
   **  Mouse control methods
   **/
  void mousePressed() {
    if (mouseX > DRAW_WINDOW_COORDS[0] && mouseX < DRAW_WINDOW_COORDS[2] && 
      mouseY >DRAW_WINDOW_COORDS[1] && mouseY < DRAW_WINDOW_COORDS[3]) {

      if (mouseEvent.getClickCount()==2) {
        if (fileStatus) {
          scaler = 1;
          nmx = nmy = nmz = 0;
          u = v = 0;
          autoState = 4;
        }
      } 
      else {
        locked = true;
        xOffset = mouseX-nmx; 
        yOffset = mouseY-nmy;
        ejxOffset = mouseX; 
        ejyOffset = mouseY;
      }
    } 
    else {
      locked = false;
    }
  }

  void mouseDragged() {

    if (locked) {
      if (fileStatus) {
        if (keyPressed) {
          v += (mouseX - ejxOffset)*0.0005;
          if ( v > TWO_PI) { 
            v = 0.0;
          }
          u -= (mouseY - ejyOffset)*0.0005;
          if ( u > TWO_PI) { 
            u = 0.0;
          }
        } 
        else {
          nmx = mouseX-xOffset; 
          nmy = mouseY-yOffset;
        }
        autoState = 4;
      }
    }
  }

  void mouseReleased() {
    locked = false;
  }

  void mouseWheel(int delta) {
    if (delta == -1) {
      scaler += 0.1;
    } // smaller
    if (delta == 1) {
      scaler -= 0.1;
    } // bigger
    if (mousePressed) {
      scaler = 1;
    } // reset scale

    scaler = constrain(scaler, 1.0, 8.0);
    autoState = 4;  // Redraw
  }

