/*
Copyright (c) 2013  V.Ayllón.  All rights reserved.
 
 DiyouPCBRun Console is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 DiyouPCBRun, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with DiyouPCBRun so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import controlP5.*;

ControlP5 cp5;
String FEED_HOLD_COMMAND = "!";
String CYCLE_START_COMMAND = "|";
String RESET_COMMAND = "\030"; // ctrl-x
//String INIT_COMMAND = "PU;IN;IP 0,0,100,100; SC 0,100,0,100;";
String LASER_POWER_COMMAND = "@P ";
int CMD_WINDOW_LENGTH = 125;
int MSG_WINDOW_WIDTH = 50;
int MAX_LASER_POWER = 255;
int MIN_SPEED = 10;
int MAX_SPEED = 250;
int DEFAULT_SPEED = 10;
int DEFAULT_LASER_POWER = 88;

boolean portStatus = false;  // false = no port
boolean connectStatus = false;  // false = disconected
boolean fileStatus = false;  // false = no loaded
int printStatus = 0;  // 0: Stopped, 1: Printing 2: Paused

int portNum = 0;

int lastLaserValue = 0;
int lastFocusValue = 0;
int lastSpeedValue = 0;

Button resetCtrl, loadFileCtrl, pauseCtrl, printCtrl;
Toggle connectCtrl;
Slider focusCtrl;
ListBox portCtrl, baudCtrl;
Textarea cmdWindow, msgWindow;
Knob laserPowerCtrl, speedCtrl;

PImage[] imgs = new PImage[16];
int BaudRate = 115200;

/**
 ** GUI initialization
 **/
void GUIinit() {


  imgs[0] = loadImage("data/print0.jpg");
  imgs[1] = loadImage("data/cancel.jpg");
  imgs[2] = loadImage("data/continue.jpg");
  imgs[3] = loadImage("data/pause0.jpg");
  imgs[4] = loadImage("data/connect0.jpg");
  imgs[5] = loadImage("data/disconnect.jpg");
  imgs[6] = loadImage("data/loadfile0.jpg");
  imgs[7] = loadImage("data/reset0.jpg");
  imgs[8] = loadImage("data/print.jpg");
  imgs[9] = loadImage("data/pause.jpg");
  imgs[10] = loadImage("data/loadfile.jpg");
  imgs[11] = loadImage("data/reset.jpg");
  imgs[12] = loadImage("data/connect.jpg");

  // The connect button

  connectCtrl = cp5.addToggle("Connect")
    .setPosition(X_MARGIN, Y_MARGIN)
      .setSize(100, 20)
        .setImage(imgs[4]);

  // The load file button

  loadFileCtrl = cp5.addButton("LoadFile")
    .setPosition(X_MARGIN+105, Y_MARGIN)
      .setSize(100, 20)
        .setImage(imgs[6]);

  // The print button

  printCtrl = cp5.addButton("StartPrint")
    .setPosition(X_MARGIN+105*2, Y_MARGIN)
      .setSize(100, 20)
        .setImage(imgs[0])
          ;  


  // The pause button

  pauseCtrl = cp5.addButton("Pause")
    .setPosition(X_MARGIN+105*3, Y_MARGIN)
      .setSize(100, 20)
        .setImage(imgs[3])
          ;

  // The reset button

  resetCtrl = cp5.addButton("Reset")
    .setPosition(X_MARGIN+105*4, Y_MARGIN)
      .setSize(100, 20)
        .setImage(imgs[7])
          ;

   // The ports list box

  portCtrl = cp5.addListBox("Ports")
    .setPosition(X_MARGIN, 80)
      .setSize(70, 30)
        .setItemHeight(15)
          .setBarHeight(15)
            .setColorBackground(color(40, 128))
              .setColorActive(color(255, 128))
                ;
  portCtrl.captionLabel().toUpperCase(true);
  portCtrl.captionLabel().set("Ports");
  portCtrl.captionLabel().setColor(0xffff0000);
  portCtrl.captionLabel().style().marginTop = 3;
  portCtrl.valueLabel().style().marginTop = 3;

  updatePortList();

 
  // The Text areas

  cmdWindow = cp5.addTextarea("Command")
    .setPosition(X_MAX_LENGTH+(X_MARGIN*2)+10, DRAW_WINDOW_COORDS[1])
      .setSize(CMD_WINDOW_LENGTH, DRAW_WINDOW_COORDS[3]-120)
        .setFont(createFont("arial", 12))
          .setLineHeight(14)
            .setColor(color(128))
              .setColorBackground(0)
                .setColorForeground(color(255, 100))
                  ;

  msgWindow = cp5.addTextarea("Messages")
    .setPosition(5, WINDOW_WIDTH-MSG_WINDOW_WIDTH)
      .setSize(WINDOW_LENGTH-X_MARGIN*2, MSG_WINDOW_WIDTH)
        .setFont(createFont("arial", 12))
          .setLineHeight(14)
            .setColor(color(128))
              .setColorBackground(0)
                .setColorForeground(color(255, 100))
                  ;

  laserPowerCtrl = cp5.addKnob("LaserPower")
    .setLabel("Laser Power")
      .setRange(0, 255)
        .setValue(DEFAULT_LASER_POWER)
          .setPosition(100, 45)
            .setRadius(30)
              .setNumberOfTickMarks(20)
                .setTickMarkLength(2)
                  .snapToTickMarks(false)
                    .setColorForeground(color(255))
                      .setColorBackground(color(221, 0, 221))
                        .setColorActive(color(255, 255, 0))
                          .setDragDirection(Knob.HORIZONTAL)
                            ;

}
/**
 ** sendBuffer() - Read next line from file and send it to serial line
 **/
String sendBuffer() {
  String buffer;

  if (rowNum < fileRow.length) {
    buffer = fileRow[rowNum];
    myPort.write(buffer);
    myPort.write("\n");
    printlnCmd(buffer);
    rowNum++;
    return buffer;
  } 
  else {
    // Stop reading because of an error or file is empty
    printlnMsg("-- End of job --");
    printlnCmd("-- End of file --");
    printCtrl.setValue(0);  // Stop button
    return null;
  }
}

/**
 ** controlEvent()
 **/
void controlEvent(ControlEvent theEvent) {
  if (theEvent.isGroup()) {
    // an event from a group e.g. scrollList
    println(theEvent.group().value()+" from "+theEvent.group());
  }

  if (theEvent.isGroup() && theEvent.name().equals("Ports")) {
    portNum = (int)theEvent.group().value();
  } 
}

/**
 ** LoadFile()
 **/
public void LoadFile(int v) {
  if (v == 1) {
    if (connectStatus && printStatus == 0) {
      loadPath = null;
      selectInput("Select a HPGL file","fileSelected");  // Opens file chooser
      autoState = 5;  // Waiting file name
    }
  }
}
/**
 ** fileSelected() - Callback fdr selectInput
 **/
void fileSelected(File selection) {
  if (selection == null) {
    loadPath = null;
  } else {
    loadPath = selection.getAbsolutePath();
  }
}
/**
 ** Connect() with printer
 **/
public void Connect(boolean f) {

  if (portStatus) {
    if (f==false) {
      connectCtrl.setImage(imgs[12]);
      myPort.stop(); // Close the port
      connectStatus = false;
      resetThis(0);
      printlnMsg("Printer disconnected");
    } 
    else {

      try {
        myPort = new Serial(this, Serial.list()[portNum], BaudRate);
      } 
      catch (Exception e) {
        printlnMsg("Error trying to connect to serial port: "+Serial.list()[portNum]);       // Error
        updatePortList();
        return;
      }
      myPort.clear();
      connectCtrl.setImage(imgs[5]);
      printlnMsg("Connected to "+Serial.list()[portNum]+" @ "+BaudRate);
      connectStatus = true;
      loadFileCtrl.setImage(imgs[10]);
      resetCtrl.setImage(imgs[11]);
      cmdWindow.clear();
    }
  } 
  else {
    updatePortList();
  }
}
/**
 ** Print() 
 **/
void StartPrint(int v) {
  if (fileStatus) {
    if (v == 1) {
      switch(printStatus) {
      case 0:   // Stopped, start printing
        cmdWindow.clear();
        printlnMsg("Printing...");
        printStatus = autoState = 1;
        rowNum = 0;  // Reset buffer #
        myColor = blueRay;  // Set the color to blue
        printCtrl.setImage(imgs[1]);
        pauseCtrl.setImage(imgs[9]);
        resetCtrl.setImage(imgs[7]);
        loadFileCtrl.setImage(imgs[6]);
        Float lp = laserPowerCtrl.getValue();
        laserPowerCtrl.setValue(0);
        laserPowerCtrl.setValue(lp);
        break;
      case 1: // Stopped
        printStatus = autoState = 0;
        printlnMsg("Printing stopped.");
        printCtrl.setImage(imgs[8]);
        loadFileCtrl.setImage(imgs[10]);
        pauseCtrl.setImage(imgs[3]);
        resetCtrl.setImage(imgs[11]);
        reopenFile();
        break;
      case 2: // Paused, continue
        printlnMsg("Printing...");
        printStatus = autoState = 1;
        printCtrl.setImage(imgs[1]);
        pauseCtrl.setImage(imgs[9]);
        resetCtrl.setImage(imgs[11]);
        myPort.write(CYCLE_START_COMMAND);
        myPort.write("\n");
        break;
      }
    }
  }
}
/**
 ** Pause()
 **/
public void Pause(int v) {

  if (v==1) {
    if (printStatus==1) {
      printCtrl.setImage(imgs[2]);
      pauseCtrl.setImage(imgs[3]);
      printStatus = 2;
      autoState = 0;
      myPort.write(FEED_HOLD_COMMAND);
      myPort.write("\n");
      printlnMsg("Printing paused.");
    }
  }
}

/**
 ** LaserPower knob
 **/
void LaserPower(int v) {
  if (v != lastLaserValue) {
    lastLaserValue = v;
    if (v >= 0 && v <= 255) {
//      int power = round((MAX_LASER_POWER*v)/100);
      buffer = LASER_POWER_COMMAND + v + ";\n";
      if (connectStatus && myPort != null) {
          myPort.write(buffer);
          myPort.write("\n");
        printlnCmd(buffer);
      }
    }
  }
}

/**
 ** Reset() - Reset the printer
 **/
public void Reset(int v) {
  if (v == 1) {
    if (connectStatus) {
      resetThis(1);
      resetPrinter();
    }
  }
}
/**
 ** resetThis() - Reset this program, not the printer
 **/
public void resetThis(int level) {
  if (fileRow != null)
    closeFile();
  printStatus = autoState = rowNum = 0;
  printCtrl.setImage(imgs[0]);
  pauseCtrl.setImage(imgs[3]);
  if (level == 0) {
    fileStatus = false;
    loadFileCtrl.setImage(imgs[6]);
      resetCtrl.setImage(imgs[7]);
  } else {
    resetCtrl.setImage(imgs[11]);
  }
}
/**
 ** updatePortList()
 **/
void updatePortList() {
  portCtrl.clear();

  if (Serial.list() != null && Serial.list().length > 0) {
    for (int i=0; i < Serial.list().length;i++) {
      ListBoxItem lbi = portCtrl.addItem(Serial.list()[i], i);
      lbi.setColorBackground(0xffff0000);
    }
    portCtrl.setValue(0);
    portStatus = true;
    connectCtrl.setImage(imgs[12]);
  } 
  else {
    portStatus = false;
  }
}
/**
 ** reopenFile()
 **/
boolean reopenFile() {
  return openFile();
}

/**
 ** openFile()
 **/
boolean openFile() {
  // If a file was selected, print path to file
  if (loadPath != null) {
    if (fileRow !=  null)
      closeFile();
    try {
      //      fileReader = createReader(loadPath);
      fileRow = loadStrings(loadPath);
    } 
    catch (Exception e) {
      fileStatus = false;
      printlnMsg("Error opening file.");
      return true;
    }
    if (fileRow ==  null) {
      fileStatus = false;
      printlnMsg("Error opening file.");
      return true;
    }
    fileStatus = true;
    printCtrl.setImage(imgs[8]);
    printlnMsg("File " + loadPath + " opened: "+ fileRow.length + " rows");
    return false;
  }
  fileStatus = false;
  printlnMsg("Error opening file.");
  return true;
}
/**
 ** closeFile()
 **/
void closeFile() {
  //  fileReader = null;
  fileRow = null;
  printCtrl.setImage(imgs[0]);
}
/**
 ** resetPrinter() - Sends a reset command to printer
 **/
void resetPrinter() {
  if (myPort != null) {
    myPort.write(RESET_COMMAND);
    myPort.write("\n");
  }
  printlnMsg("Printer reset");
}
/**
 * Print a message in the message window
 **/
void printMsg(String s) {
  msgWindow.setText(s);
}
/**
 * Print a message in the message window
 **/
void printlnMsg(String s) {
  msgWindow.setText(s+"\n");
}
/**
 * Print a command in the command window
 **/
void printlnCmd(String s) {
  cmdWindow.append(s+"\n");
  cmdWindow.scrolled(5);
}


/**
 ** clearDraw - clear the draw
 
 void clearDraw() {
 pg.beginDraw();
 pg.rectMode(CORNERS);
 
 pg.fill(0);
 pg.stroke(255, 255, 255);
 pg.rect(0, 0, DRAW_WINDOW_COORDS[2], DRAW_WINDOW_COORDS[3]);
 pg.endDraw();
 imageMode(CORNERS);
 image(pg, DRAW_WINDOW_COORDS[0], DRAW_WINDOW_COORDS[1], DRAW_WINDOW_COORDS[2], DRAW_WINDOW_COORDS[3]);
 }
 
 **/
