/*
Copyright (c) 2013  V.Ayllón.  All rights reserved.
 
 Inspired by HP2xx:
 
 Parts Copyright (c) 1991 - 1994 Heinz W. Werntges.  All rights reserved.
 Parts Copyright (c) 1998-2002,2003 Martin Kroeker  All rights reserved.
 
 DiyouPCBRun Console is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 DiyouPCBRun, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with DiyouPCBRun so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
float P1X_default = 0.0;
float P1Y_default = 0.0;
float P2X_default = 100.0;	
float P2Y_default = 100.0;

float PLOT_UNIT	= 0.025f*SCALE;

int  MAX_CMD_SIZE = 96;
int  MAX_PARAMETERS = 4;

char cmdbuffer[] = new char[255];

// PA command parameters
int  X_COORD = 0;
int  Y_COORD = 1;

// AA command parameters

int  I_COORD  = 0;
int  J_COORD  = 1;
int  ARC_THETA = 2;
int  SEG_ALPHA = 3;

// IP command parameters

int  IP_P1X = 0;
int  IP_P1Y = 1;
int  IP_P2X = 2;
int  IP_P2Y = 3;

// SC command parameters

int  SC_P1X=0;
int  SC_P1Y=1;
int  SC_P2X=2;
int  SC_P2Y=3;


/* Supported HPGL commands, ASCII-coded as High-byte/low-byte int's */

int  AA	= 0x4141;
int  IN	= 0x494E;
int  IP	= 0x4950;
int  PA	= 0x5041;
int  PD	= 0x5044;
int  PU = 0x5055;
int  SC	= 0x5343;
int  SP	= 0x5350;
int  VS	= 0x5653;
int  PW	= 0x5057;
// Command error
int  PARSER_ERROR = 0xffff;


float[] numpar = new float[MAX_PARAMETERS];
char[] buffptr;
int buffinx = 0;

float[] P1 = new float[NUM_AXIS];	// Scaling points
float[] P2 = new float[NUM_AXIS];
float[] S1 = new float[NUM_AXIS];
float[] S2 = new float[NUM_AXIS];
float[] Q = new float[NUM_AXIS];

/** HPGLInit() - Init variables
 ** HP-GL commands, reads parameters (if expected), and initiates processing of these
 ** commands.
 **/
void HPGLInit() {

  P1[X_COORD] =  P1X_default;
  P1[Y_COORD] =  P1Y_default;
  P2[X_COORD] =  P2X_default; 
  P2[Y_COORD] =  P2Y_default;

  S1[X_COORD] =  P1X_default;
  S1[Y_COORD] =  P1Y_default;
  S2[X_COORD] =  P2X_default; 
  S2[Y_COORD] =  P2Y_default;


  Q[X_COORD] = 1.; 		/* Delta-P/Delta-S: Initialized with first SC   */
  Q[Y_COORD] = 1.;
}
/** HPGLParserRun() - Reads the input stream character-by-character, identifies
 ** HP-GL commands, reads parameters (if expected), and initiates processing of these
 ** commands.
 **/
int HPGLParserRun(char[] buffer) {
  buffptr = buffer;
  buffinx = 0;
  int c;
  int cmd;

  while ( (c = buffer[buffinx++]) != '\0') {
    if ((c < '@') || (c > 'z')
      || ((c > 'Z') && (c < 'a')))
      continue;
    cmd = c << 8;
    if ((c = buffptr[buffinx++]) == '\0')
      break;	// Partial command
    if ((c < '@') || (c > 'z')
      || ((c > 'Z') && (c < 'a'))) {
      continue;
    }
    cmd |= (c & 0xFF);
    cmd &= 0xDFDF; // Pass to upper
    //			print(pbuffer);
    return HPGLParseCommand(cmd);
  }
  return PARSER_ERROR;
}

/**
 ** HPGLParseCommand() - each command consists of 2 characters
 ** This function returns the command detected.
 **/
int HPGLParseCommand(int command) {

  //			print(pbuffer);

  switch (command) {	
  case 0x4141: //AA:
    if (getAACoordinates() > 0)
      return PARSER_ERROR;
    break;
  case 0x5041: //PA:
    if (getPACoordinates() > 0)
      return PARSER_ERROR;
    break;
  case 0x4048: //@H
  case 0x4049: //@I
  case 0x5044: //PD:
  case 0x5350: //SP:		// Select pen 
  case 0x5055: //PU:
    break;
  case 0x4950: //IP:		// Init point
    if (getIPCoordinates() > 0)
      return PARSER_ERROR;
    break;
  case 0x5343: //SC:		// Scale all of them processed inside parser
    if (getSCCoordinates() > 0)
      return PARSER_ERROR;
    break;
  case 0x494E: //IN:		// Home & init
    break;
  case 0x5653: //VS:
  case 0x5057: //PW:
    if (getVSValue() > 0)
      return PARSER_ERROR;
      break;
  default: 		/* Skip unknown HPGL command: */
    break;
  }

  return command;
}


/**
 ** getIPCoordinates() - Gets IP parameters
 **
 **/
int getIPCoordinates() {
  if ((numpar[IP_P1X] = getFloat())==PARSER_ERROR) {	     
    P1[X_COORD] = P1X_default;
    P1[Y_COORD] = P1Y_default;
    P2[X_COORD] = P2X_default;
    P2[Y_COORD] = P2Y_default;

    // Set S1
    S1[X_COORD] = P1[X_COORD];
    S1[Y_COORD] = P1[Y_COORD];
    S2[X_COORD] = P2[X_COORD];
    S2[Y_COORD] = P2[Y_COORD];

    Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
    Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

    return 0;
  }
  if ((numpar[IP_P1Y] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR; /* x without y! */

  if ((numpar[IP_P2X] = getFloat()) ==PARSER_ERROR) { 	/* No number found  */
    P2[X_COORD] += numpar[IP_P1X] - P1[X_COORD];
    P2[Y_COORD] += numpar[IP_P1Y] - P1[Y_COORD];

    P1[X_COORD] = numpar[IP_P1X];
    P1[Y_COORD] = numpar[IP_P1Y];
    // Set S1
    S1[X_COORD] = P1[X_COORD];
    S1[Y_COORD] = P1[Y_COORD];
    S2[X_COORD] = P2[X_COORD];
    S2[Y_COORD] = P2[Y_COORD];

    Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
    Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

    return 0;
  }
  if ((numpar[IP_P2Y] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR; /* x without y! */

  P1[X_COORD] = numpar[IP_P1X];
  P1[Y_COORD] = numpar[IP_P1Y];
  P2[X_COORD] = numpar[IP_P2X];
  P2[Y_COORD] = numpar[IP_P2Y];

  // Set S1
  S1[X_COORD] = P1[X_COORD];
  S1[Y_COORD] = P1[Y_COORD];
  S2[X_COORD] = P2[X_COORD];
  S2[Y_COORD] = P2[Y_COORD];

  Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
  Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

  return 0;
}

/**
 ** getSCCoordinates() - Gets SC parameters
 **
 **/
int getSCCoordinates() {

  if ((numpar[SC_P1X] = getFloat())==PARSER_ERROR) {	     
    S1[X_COORD] = P1X_default;
    S1[Y_COORD] = P1Y_default;
    S2[X_COORD] = P2X_default;
    S2[Y_COORD] = P2Y_default;
    Q[X_COORD] = Q[Y_COORD] = 1.0;
    return 0;
  }
  if ((numpar[SC_P2X] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR; /* x without y! */

  if ((numpar[SC_P1Y] = getFloat()) ==PARSER_ERROR) { 	/* No number found  */
    return PARSER_ERROR; /* x without y! */
  }
  if ((numpar[SC_P2Y] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR; /* x without y! */


  if (numpar[SC_P1X] == numpar[SC_P2X] || numpar[SC_P1Y] == numpar[SC_P2Y]) { 	/* min must differ from max */
    Q[X_COORD] = Q[Y_COORD] = 1.0;
    return 0;
  }

  // Set S1

  S1[X_COORD] = numpar[SC_P1X];
  S1[Y_COORD] = numpar[SC_P1Y];

  S2[X_COORD] = numpar[SC_P2X];
  S2[Y_COORD] = numpar[SC_P2Y];

  // Update Delta

  Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
  Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

  return 0;
}


/**
 ** getVSValue() - Gets the value of a speed command
 **/
int getVSValue() {
  if ((numpar[0] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR;
  return 0;
}
/**
 ** getPACoordinates() - Gets the coordinates of a move command
 **/
int getPACoordinates()
{

  if ((numpar[X_COORD] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR;
  if ((numpar[Y_COORD] = getFloat())==PARSER_ERROR)	      
    return PARSER_ERROR;
  /**
   println("\ngetPACoordinates() - BEF");
   print(" numpar X: ");
   print(numpar[0]);
   print(" numpar Y ");
   println(numpar[1]);
   **/
  numpar[X_COORD] = HPGLPlotterToUSer(numpar[X_COORD], X_COORD);
  numpar[Y_COORD] = HPGLPlotterToUSer(numpar[Y_COORD], Y_COORD);

  return 0;
}
/**
 * getAACoordinates() - Gets the arc command coordinates
 **/
int getAACoordinates()
{
  if ((numpar[I_COORD] = getFloat())==PARSER_ERROR)	     
    return PARSER_ERROR;
  if ((numpar[J_COORD] = getFloat())==PARSER_ERROR)	      
    return PARSER_ERROR;
  if ((numpar[ARC_THETA] = getFloat())==PARSER_ERROR)	      
    return PARSER_ERROR;
  numpar[SEG_ALPHA] = getFloat();	// Opcional


  numpar[I_COORD] = HPGLPlotterToUSer(numpar[I_COORD], I_COORD);
  numpar[J_COORD] = HPGLPlotterToUSer(numpar[J_COORD], J_COORD);
/**
  // Adjust 360 arcs
  if (numpar[ARC_THETA] == 360) {
    numpar[ARC_THETA] = numpar[ARC_THETA] - .001f;
  }
  else if (numpar[ARC_THETA] == -360) {
    numpar[ARC_THETA] = numpar[ARC_THETA] + .001f;
  }
  **/
  return 0;
}


/**
 ** getFloat() - only accepts float positive numbers
 ** returns PARSER_ERROR if error
 **/
float getFloat()
{
  char c;
  int n = 0;
  char[] numbuf =  new char[80];
  //        char numbuf[80];


  while ( (c = buffptr[buffinx++]) != '\0') {
    if (c == ' ' || c == '\t' ||c == '\n' || c== '\r')
      continue;
    if (((c >= '0') && (c <= '9')) || (c == '.') || (c == '+') || (c == '-')) {
      numbuf[n++] = c;
    } 
    else { // end of number
      numbuf[n] = '\0';
      return (float)atof(numbuf);
    }
  }
  return PARSER_ERROR;
}

/**
 ** HPGLUserToPlotter() - Translate user to plotter coordinates
 ** Is so basic in order to maintain compatibility with Processing.
 **/
float HPGLUserToPlotter(float coord, int axis)
{
  if (axis > NUM_AXIS)
    return -1;
  return (float)(P1[axis] + (coord - S1[axis]) * Q[axis])/PLOT_UNIT;
}
/**
 ** HPGLPlotterToUSer() - Translate plotter to user coordinates
 ** Is so basic in order to maintain compatibility with Processing.
 **/
float HPGLPlotterToUSer(float coord, int axis)
{
  if (axis > NUM_AXIS)
    return -1;
  return  (float)(S1[axis] + (coord - P1[axis])/Q[axis])*PLOT_UNIT;
}
/**
 * atof() - Processing doesn´t have.
 **/
float atof( char[] s) { 
  String str2 = new String(s);

  return float(str2);
}

