/*
Copyright (c) 2013  V.Ayllón.  All rights reserved.
 
 Inpired by GRBL:
 
 Copyright (c) 2009-2011 Simen Svale Skogsrud
 Copyright (c) 2011 Sungeun K. Jeon
 
 
 DiyouPCBRun Console is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 DiyouPCBRun, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with DiyouPCBRun so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
float MM_PER_ARC_SEGMENT_SMALL = .1;
float SMALL_CIRCLE_RADIUS = 2.5;
float MM_PER_ARC_SEGMENT_LARGE = 1;

int N_ARC_CORRECTION = 25;


/**
 ** homeAllAxis() - Home all axis
 **/
void homeAllAxis() {
  turnLaser(OFF);
  current_position[X_AXIS] = 0.f;
  current_position[Y_AXIS]  = Y_MAX_LENGTH;
  destination[X_AXIS] = 0.f;
  destination[Y_AXIS] = Y_MAX_LENGTH;
}


/**
 ** setFeedRate() - Set the max. feedrate
 **/

void setFeedRate(float frate) {
  //  print("\nsetFeedRate: ");
  //  println(frate);
}
/**
 ** setCoordinates() - Sets the coordinates of a move command
 **/
void setCoordinates()
{

  destination[X_AXIS] = (float)numpar[0]; // + current_position[X_AXIS];
  destination[Y_AXIS] = (float)numpar[1]; // + current_position[Y_AXIS];
}
/**
 * prepareMove() - Prepare the movement
 **/
void prepareMove()
{

  //  planBufferLine(destination[X_AXIS], destination[Y_AXIS], feedrate*feedmultiply/60/100.0,laser_state);
  if (laser_state) {
    pg.stroke(myColor);  // Blue ray
    pg.line(current_position[X_AXIS], current_position[Y_AXIS], destination[X_AXIS], destination[Y_AXIS]);
  }

  for (int i=0; i < NUM_AXIS; i++) {
    current_position[i] = destination[i];
  }
}
/**
 * hypot() - Calc hypotenuse
 **/
float hypot(float cat1, float cat2) {

  return sqrt(cat1*cat1+cat2*cat2);
}

/**
 * prepareArcMove() - Prepare the movement of an arc command
 **/
void prepareArcMove() {
  float theta;
  boolean isClockwise;

  theta = radians(numpar[ARC_THETA]*-1); // ADAPTED to screen origin

  destination[X_AXIS] = numpar[I_COORD] + (current_position[X_AXIS] - numpar[I_COORD]) * cos(theta) - (current_position[Y_AXIS] - numpar[J_COORD]) * sin(theta);
  destination[Y_AXIS] = numpar[J_COORD] + (current_position[X_AXIS] - numpar[I_COORD]) * sin(theta) + (current_position[Y_AXIS] - numpar[J_COORD]) * cos(theta);

  offset[X_AXIS] = numpar[I_COORD]-current_position[X_AXIS];		// I (x of center)
  offset[Y_AXIS] = numpar[J_COORD]-current_position[Y_AXIS];		// J (y of center)

  float radius = hypot(offset[X_AXIS], offset[Y_AXIS]); // Compute radius

  if (numpar[ARC_THETA] < 0)	//Angle of rotation,ADAPTED to screen origin
    isClockwise = false;
  else
    isClockwise = true;
  // Trace the arc

  if (!mc_arc(current_position, destination, offset, X_AXIS, Y_AXIS, feedrate*feedmultiply/60/100.0, radius, isClockwise))
    return;
  // If no error update position
  for (int i=0; i < NUM_AXIS; i++) {
    current_position[i] = destination[i];
  }

  // As far as the parser is concerned, the position is now == target. In reality the
  // motion control system might still be processing the action and the real tool position
  // in any intermediate location.
}

// The arc is approximated by generating a huge number of tiny, linear segments. The length of each 
// segment is configured in settings.mm_per_arc_segment.  
boolean mc_arc(float[] position, float[] target, float[] offset, int axis_0, int axis_1, float feed_rate, float radius, boolean isclockwise)
{      
  float currentx =   position[axis_0];
  float currenty =   position[axis_1];
  //   int acceleration_manager_was_enabled = plan_is_acceleration_manager_enabled();
  //   plan_set_acceleration_manager_enabled(false); // disable acceleration management for the duration of the arc
  float center_axis0 = position[axis_0] + offset[axis_0];
  float center_axis1 = position[axis_1] + offset[axis_1];
  float r_axis0 = -offset[axis_0];  // Radius vector from center to current location
  float r_axis1 = -offset[axis_1];
  float rt_axis0 = target[axis_0] - center_axis0;
  float rt_axis1 = target[axis_1] - center_axis1;
  int segments;
  float angular_travel;

  // CCW angle between position and target from circle center. Only one atan2() trig computation required.
  angular_travel = atan2(r_axis0*rt_axis1-r_axis1*rt_axis0, r_axis0*rt_axis0+r_axis1*rt_axis1);

  if (isclockwise) { // Correct atan2 output per direction
    if (angular_travel >= 0) { 
      angular_travel -= 2*PI;
    }
  } 
  else {
    if (angular_travel <= 0) { 
      angular_travel += 2*PI;
    }
  }

  float millimeters_of_travel = hypot(angular_travel*radius, 0.0);     

  if (millimeters_of_travel < 0.001) { 
    return false;
  }

  if (radius < SMALL_CIRCLE_RADIUS)
    segments = floor(millimeters_of_travel/MM_PER_ARC_SEGMENT_SMALL);
  else
    segments = floor(millimeters_of_travel/MM_PER_ARC_SEGMENT_LARGE);

  if (segments == 0) segments = 1;

  /*  
   // Multiply inverse feed_rate to compensate for the fact that this movement is approximated
   // by a number of discrete segments. The inverse feed_rate should be correct for the sum of 
   // all segments.
   if (invert_feed_rate) { feed_rate *= segments; }
   */
  float theta_per_segment = angular_travel/segments;


  /* Vector rotation by transformation matrix: r is the original vector, r_T is the rotated vector,
   and phi is the angle of rotation. Based on the solution approach by Jens Geisler.
   r_T = [cos(phi) -sin(phi);
   sin(phi)  cos(phi] * r ;
   
   For arc generation, the center of the circle is the axis of rotation and the radius vector is 
   defined from the circle center to the initial position. Each line segment is formed by successive
   vector rotations. This requires only two cos() and sin() computations to form the rotation
   matrix for the duration of the entire arc. Error may accumulate from numerical round-off, since
   all double numbers are single precision on the Arduino. (True double precision will not have
   round off issues for CNC applications.) Single precision error can accumulate to be greater than
   tool precision in some cases. Therefore, arc path correction is implemented. 
   
   Small angle approximation may be used to reduce computation overhead further. This approximation
   holds for everything, but very small circles and large mm_per_arc_segment values. In other words,
   theta_per_segment would need to be greater than 0.1 rad and N_ARC_CORRECTION would need to be large
   to cause an appreciable drift error. N_ARC_CORRECTION~=25 is more than small enough to correct for 
   numerical drift error. N_ARC_CORRECTION may be on the order a hundred(s) before error becomes an
   issue for CNC machines with the single precision Arduino calculations.
   
   This approximation also allows mc_arc to immediately insert a line segment into the planner 
   without the initial overhead of computing cos() or sin(). By the time the arc needs to be applied
   a correction, the planner should have caught up to the lag caused by the initial mc_arc overhead. 
   This is important when there are successive arc motions. 
   */
  // Vector rotation matrix values
  float cos_T = 1-0.5*theta_per_segment*theta_per_segment; // Small angle approximation
  float sin_T = theta_per_segment;

  float[] arc_target = new float[2];
  float sin_Ti;
  float cos_Ti;
  float r_axisi;
  int i;
  int count = 0;

  for (i = 1; i<segments; i++) { // Increment (segments-1)

    if (count < N_ARC_CORRECTION) {
      // Apply vector rotation matrix 
      r_axisi = r_axis0*sin_T + r_axis1*cos_T;
      r_axis0 = r_axis0*cos_T - r_axis1*sin_T;
      r_axis1 = r_axisi;
      count++;
    } 
    else {
      // Arc correction to radius vector. Computed only every N_ARC_CORRECTION increments.
      // Compute exact location by applying transformation matrix from initial radius vector(=-offset).
      cos_Ti = cos(i*theta_per_segment);
      sin_Ti = sin(i*theta_per_segment);
      r_axis0 = -offset[axis_0]*cos_Ti + offset[axis_1]*sin_Ti;
      r_axis1 = -offset[axis_0]*sin_Ti - offset[axis_1]*cos_Ti;
      count = 0;
    }

    // Update arc_target location
    arc_target[axis_0] = center_axis0 + r_axis0;
    arc_target[axis_1] = center_axis1 + r_axis1;
    //    clamToSoftwareEnstops(arc_target);
    if (laser_state) {
      pg.stroke(myColor);  // Blue ray
      pg.line(currentx, currenty, arc_target[axis_0], arc_target[axis_1]);
      currentx = arc_target[axis_0];
      currenty = arc_target[axis_1];
    }
  }
  // Ensure last segment arrives at target location.
  if (laser_state) {
    pg.stroke(myColor);  // Blue ray
    pg.line(currentx, currenty, target[axis_0], target[axis_1]);
  }

  return true;
}
/**
 * clampToSofwareEndstops() 
 **/
void clamToSoftwareEnstops(float[] target)
{

  if (target[X_AXIS] < X_MIN_PRT) target[X_AXIS] = X_MIN_PRT;
  if (target[Y_AXIS] < Y_MIN_PRT) target[Y_AXIS] = Y_MIN_PRT;

  if (target[X_AXIS] > X_MAX_PRT) target[X_AXIS] = X_MAX_PRT;
  if (target[Y_AXIS] > Y_MAX_PRT) target[Y_AXIS] = Y_MAX_PRT;
}

/**
 ** worldMap Histogram
 **/
void cleanScreen() {

  pg.rectMode(CORNER);
  pg.fill(0);
  pg.stroke(0, 0, 0);
  pg.rect(0, 0, X_MAX_LENGTH, Y_MAX_LENGTH);
}
/**
 * turnLaser() - Turn off the laser
 * Used only in case of emergency and in AutoFocus
 **/
void turnLaser(boolean state)
{
  laser_state = state;
}

